"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.textpack = void 0;
var seedrandom_1 = __importDefault(require("seedrandom"));
var regex = /{([^{}]+)}/g;
var parentSymbol = Symbol('parent');
function findInScope(scope, key) {
    while (scope) {
        if (scope.hasOwnProperty(key)) {
            return scope[key];
        }
        scope = scope[parentSymbol];
    }
    return null;
}
function getWeight(part) {
    if (typeof part === 'object' && !part.hasOwnProperty('_') && part.hasOwnProperty('weight')) {
        // @ts-ignore
        return part.weight;
    }
    return 1;
}
function evaluate(expr, config, args, source) {
    if (typeof config === 'object' && config.strict && !expr) {
        throw new (ReferenceError.bind.apply(ReferenceError, __spreadArrays([void 0], (source ? [source] : []))))();
    }
    else if (!expr) {
        return source;
    }
    return expr(args, source);
}
function resolveConfig(config) {
    var rng;
    if (typeof config === 'string') {
        rng = seedrandom_1.default(config);
        config = { rng: rng };
    }
    else if (!config.rng || typeof config.rng === 'string') {
        rng = seedrandom_1.default(config.rng);
        config = __assign(__assign({}, config), { rng: rng });
    }
    else {
        rng = config.rng;
    }
    return { config: config, rng: rng };
}
function compile(part, config, scope) {
    var _a;
    if (typeof part === 'function') {
        return part;
    }
    var resolved = resolveConfig(config);
    config = resolved.config;
    var rng = resolved.rng;
    if (typeof part === 'string') {
        var str_1 = part.trim();
        return function (args) {
            var _a;
            var argScope = (_a = {}, _a[parentSymbol] = scope, _a);
            for (var i = 0; i < args.length; i++) {
                argScope[i] = compile(args[i], config, argScope);
            }
            return str_1.replace(regex, function (match, key, index) {
                if (str_1[index - 1] === '{' && str_1[index + match.length] === '}') {
                    return key;
                }
                else {
                    var values = key.split(' ')
                        .map(function (k) { return k.trim(); })
                        .filter(function (k) { return k; });
                    var fn = findInScope(argScope, values[0]);
                    var args_1 = values.slice(1).map(function (k) {
                        var val = findInScope(argScope, k);
                        return evaluate(val, config, [], match);
                    });
                    return evaluate(fn, config, args_1, match);
                }
            });
        };
    }
    else if (Array.isArray(part)) {
        if (part.length > 0) {
            var weights = part.map(getWeight);
            var sum = 0;
            for (var i = 0; i < weights.length; i++) {
                sum += weights[i];
            }
            var random = rng() * sum;
            for (var i = 0; i < weights.length; i++) {
                var weight = weights[i];
                if (random < weight) {
                    return compile(part[i], config, scope);
                }
                random -= weight;
            }
        }
    }
    else if (part) {
        /*if(part.hasOwnProperty('_')) {
            scope = {[parentSymbol]: scope};
            for(let [k, v] of Object.entries(part)) {
                scope[k] = compile(v, rng, scope);
            }
            return scope._;
        }
        else */
        if (part.hasOwnProperty('value')) {
            if (part.scope) {
                scope = (_a = {}, _a[parentSymbol] = scope, _a);
                if (part.scope) {
                    for (var _i = 0, _b = Object.entries(part.scope); _i < _b.length; _i++) {
                        var _c = _b[_i], k = _c[0], v = _c[1];
                        scope[k] = compile(v, config, scope);
                    }
                }
            }
            if (part.params) {
                return function (args, source) {
                    var _a;
                    var argScope = (_a = {}, _a[parentSymbol] = scope, _a);
                    var params = part.params;
                    var _loop_1 = function (i) {
                        argScope[params[i]] = function () { return args[i]; };
                    };
                    for (var i = 0; i < params.length && i < args.length; i++) {
                        _loop_1(i);
                    }
                    var fn = compile(part.value, config, argScope);
                    return evaluate(fn, config, args, source);
                };
            }
            return compile(part.value, config, scope);
        }
    }
    return null;
}
function textpack(part, config, scope) {
    if (config === void 0) { config = {}; }
    config = resolveConfig(config).config;
    var compiledScope = {};
    if (scope) {
        for (var _i = 0, _a = Object.entries(scope); _i < _a.length; _i++) {
            var _b = _a[_i], key = _b[0], value = _b[1];
            compiledScope[key] = compile(value, config, compiledScope);
        }
    }
    var result = compile(part, config, compiledScope);
    return evaluate(result, config, [], '');
}
exports.textpack = textpack;
exports.default = textpack;
